#!/usr/bin/python2
# -*- coding: utf-8 -*-
"""
This is the main module of Synapse Labels
"""
# standard impoprts
from __future__ import absolute_import
from __future__ import print_function
import tempfile
import subprocess
import ftplib
import datetime
import os
from os.path import isfile
from os.path import join
from threading import Thread
from logging import basicConfig
from logging import INFO
# 3rd party imports
import cairo
from yaml import dump
from yaml import load
# from six import iteritems
from labeldata import save_label
from labeldata import now_string
from labeldata import LABELS_PRINTED_FILE
from labeldata import LABELS_GENERATED_FILE
from .exceptions import ImageFileMissing
from .demodata import DEMO_WITH_HEADER
from .svgparse import SvgParser
from .generator import Generator
from .imgconverter import convert_png_to_bmp
from .jscript_utils import shift_and_format_items
from .pw_utils import PasswordGenerator
from .references import MultipleFiles
from .references import MissingFile
from .references import ReferenceConfig
from .references import reference_config_file
from .labelconfig import label_config_file
from .draw import draw_label_file
from .utils import svg_dir
from .utils import get_path
from .utils import list_of_dicts_to_table
from .utils import start_file
from .utils import on_drag_data_get
from .utils import set_treeview_columns
from .utils import set_treeview_from_clipboard
from .utils import set_treeview_from_text_data
from .utils import set_treeview_from_texts
from .utils import fill_combo
from .utils import drag_data_received
from .utils import print_label_file_to

try:
    from synapselabels import __version__

    assert __version__
except ImportError:
    __version__ = "UNKNOWN"
try:
    from gi import require_version
except ImportError:
    print("ERROR: pygobject is not installed.")
    raise
try:
    require_version("Gtk", "3.0")
except ValueError:
    print("ERROR: gtk3 development files are not installed.")
    raise
except AttributeError:
    print("pygobject version too old.")
    raise
try:
    from gi.repository import Gtk
    from gi.repository import Gdk
    from gi.repository import GLib
    # from gi.repository import GObject
except ImportError:
    print("ERROR: pygobject is not installed.")
    raise

# log = getLogger('synapselabels')
GLADE_FILE = get_path('labelsgui.glade')
SRC1 = get_path('names.txt')
SRC2 = get_path('places.txt')
DND_TARGETS = [('text/plain', Gtk.TargetFlags.OTHER_WIDGET, 0), ]


class LabelsGui(SvgParser, ReferenceConfig, Generator):
    """
    This is the E20 GUI
    """
    HEADER_HEIGHT = 10 * 72 / 25.4
    HEADER_GAP = 3 * 72 / 25.4

    def __init__(self):
        """
        Build GUI
        """
        SvgParser.__init__(self, label_config_file)
        ReferenceConfig.__init__(self, reference_config_file)
        self.template = None
        self.sales_level_part_number = None
        self.qr_code_standard = None
        self.document_name = None
        self.selected_label = None
        self.selected_label_file = None
        self.dpi = 300
        # Build GUI from Glade file
        self.builder = Gtk.Builder()
        self.builder.add_from_file(GLADE_FILE)
        # Get objects
        go = self.builder.get_object
        self.window = go('applicationwindow1')
        self.window.set_title(
            'Synapse Labels ' + __version__
        )
        self.textview = go('textview')
        self.drawingarea = go('drawingarea')
        self.drawingarea.connect("draw", self.on_draw)
        self.treeview_data_table = go('treeviewDataTable')
        self.treeview_data_table.get_selection().connect(
            "changed",
            self.data_tree_selection_changed,
            "dummy"
        )
        self.treeview_data_table.enable_model_drag_dest(
            DND_TARGETS,
            Gdk.DragAction.DEFAULT | Gdk.DragAction.MOVE
        )
        self.treeview_data_table.connect(
            'drag-data-received',
            drag_data_received,
        )
        # self.combobox = go('comboboxtext1')
        # self.mac_num = go('labelMacNum')
        self.entry_date_code = go('entryDateCode')
        self.entry_records = go('entryRecords')
        self.entry_records.connect(
            "changed",
            self.entry_records_changed
        )
        self.entry_copies = go('entryCopies')
        self.label_number_of_records = go('labelNumberOfRecords')
        self.button_view_document = go('buttonViewDocument')
        self.button_paste_with_header = go('buttonPasteWithHeader')
        self.button_print_selected_labels = go('buttonPrintSelectedLabels')
        self.button_print_every_label = go('buttonPrintEveryLabel')
        self.button_sheet_selected_labels = go('buttonSheetSelectedLabels')
        self.button_sheet_every_label = go('buttonSheetEveryLabel')
        self.cb_printers = go('comboboxPrinters')
        self.cb_printers.connect(
            "changed",
            self.cb_printers_changed,
            "dummy"
        )
        self.cb_part_numbers = go('comboboxPartNumbers')
        self.treeview_params = go('treeviewParams')
        set_treeview_columns(self.treeview_params, ['Params'])
        self.treeview_params.enable_model_drag_source(
            Gdk.ModifierType.BUTTON1_MASK,
            DND_TARGETS,
            Gdk.DragAction.DEFAULT | Gdk.DragAction.MOVE
        )
        self.treeview_params.connect('drag_data_get', on_drag_data_get)
        self.button_generate = go('buttonGenerate')
        self.entry_spd = go('entrySpd')
        self.entry_document = go('entryDocument')
        self.entry_label_template = go('entryLabelTemplate')
        self.entry_label_type = go('entryLabelType')
        self.entry_qr_code_standard = go('entryQrCodeStandard')
        self.box_qr_code_standard = go('boxQrCodeStandard')
        self.box_preview = go('boxPreview')
        self.box_upgrade = go('boxUpgrade')
        fill_combo(self.cb_printers, sorted(self.printers.keys()))
        self.cb_printers.set_active(-1)
        fill_combo(self.cb_part_numbers, sorted(self.part_numbers.keys()))
        self.cb_part_numbers.set_active(-1)
        self.cb_documents = go('comboboxDocuments')
        self.cb_documents.connect(
            "changed",
            self.cb_documents_selection_changed,
            "dummy"
        )
        self.cb_part_numbers.connect(
            "changed",
            self.cb_part_numbers_selection_changed,
            "dummy"
        )
        self.treeview_params.get_selection().connect(
            "changed",
            self.params_tree_selection_changed,
            "dummy"
        )
        self.set_date_code_to_today()
        # Connect signals
        self.builder.connect_signals(self)
        self.cb_part_numbers_selection_changed(self.cb_part_numbers)
        # Configure interface
        self.window.connect('delete-event', lambda x, y: Gtk.main_quit())
        # GObject.timeout_add(1000, self.update_mac_pool)
        # Everything is ready
        self.window.show()
        self.password_generator = PasswordGenerator(SRC1, SRC2)
        thread = Thread(target=self.check_update)
        thread.daemon = True
        thread.start()

    def button_upgrade_clicked(self, widget, data=None): \
            # pylint: disable=unused-argument
        thread = Thread(target=self.upgrade)
        thread.daemon = True
        thread.start()
        self.box_upgrade.set_visible(False)

    def upgrade(self):
        args = [
            'pip2',
            'install',
            '-U',
            '--disable-pip-version-check',
            '--extra-index-url',
            'https://update.synapse-wireless.com/pypi',
            'synapselabels',
        ]
        try:
            print(subprocess.check_output(args))
            print('upgrade completed successfully, please restart')
        except subprocess.CalledProcessError:
            # sudo is not working if operator is logged in!
            # if os.name == 'posix':
            #     print('upgrade failed, retrying upgrade with gksudo')
            #     args = ['gksudo', ' '.join(args)]
            #     try:
            #         print(subprocess.check_output(args))
            #         print('upgrade completed successfully, please restart')
            #         return
            #     except subprocess.CalledProcessError:
            #         print('upgrade failed, please upgrade manually')
            #         self.box_upgrade.set_visible(True)
            print('upgrade failed, please upgrade manually')
            self.box_upgrade.set_visible(True)

    def check_update(self):
        self.box_upgrade.set_visible(False)
        self.window.set_title(
            'Synapse Labels ' + __version__ + ' (version check in progress)'
        )
        args = [
            'pip2',
            'list',
            '--disable-pip-version-check',
            '--extra-index-url',
            'https://update.synapse-wireless.com/pypi',
            '-o',
            '--format=legacy'
        ]
        try:
            updatables = subprocess.check_output(args)
            for updatable in updatables.splitlines():
                if updatable.startswith('synapselabels '):
                    latest_index = updatable.index('Latest: ')
                    latest = updatable[latest_index + 8:]
                    latest = latest.split(' ', 1)
                    latest = latest[0]
                    print('latest:', latest)
                    self.box_upgrade.set_visible(True)
                    self.window.set_title(
                        'Synapse Labels ' +
                        __version__ +
                        ' (latest version is %s)' % latest
                    )
                    return
            self.window.set_title(
                'Synapse Labels ' + __version__ + ' (this is latest version)'
            )
        except subprocess.CalledProcessError:
            self.window.set_title(
                'Synapse Labels ' + __version__ + ' (version check failed)'
            )

    def get_copies(self):
        try:
            return int(self.entry_copies.get_text())
        except ValueError:
            return 1

    def button_print_selected_labels_clicked(self, widget, data=None): \
            # pylint: disable=unused-argument
        self.print_labels(
            self.get_selected_labels(),
            self.entry_label_type.get_text(),
            self.get_copies(),
        )

    def button_sheet_selected_labels_clicked(self, widget, data=None): \
            # pylint: disable=unused-argument
        self.sheet_labels(
            self.get_selected_labels(),
            self.entry_label_type.get_text(),
            self.get_copies(),
        )

    def get_selected_labels(self):
        # print("print_selected_label")
        selection = self.treeview_data_table.get_selection()
        # print('selection:', selection)
        (model, pathlist) = selection.get_selected_rows()
        labels = []
        for path in pathlist:
            # print('path:', path)
            tree_iter = model.get_iter(path)
            # print('tree_iter:', tree_iter)
            label_data = {}
            for param in self.get_parameters():
                value = None
                for index, column in enumerate(
                        self.treeview_data_table.get_columns()):
                    if column.get_title() == param:
                        value = model.get_value(
                            tree_iter,
                            index,
                        )
                        break
                # print('value:', value)
                if value is None:
                    if param == 'date code':
                        entry_date_code_text = self.entry_date_code.get_text()
                        if len(entry_date_code_text) == 4:
                            value = entry_date_code_text
                if value:
                    label_data[param] = value
                else:
                    print(repr(param), 'parameter is missing')
                    return
            labels.append(label_data)
        return labels

    def get_label_from_row(self, row):
        label_data = {}
        for param in self.get_parameters():
            value = None
            for index, column in enumerate(
                    self.treeview_data_table.get_columns()):
                if column.get_title() == param:
                    value = row[index]
                    break
            # print('value:', value)
            if value:
                label_data[param] = value
            else:
                print(repr(param), 'parameter is missing')
                return
        return label_data

    def get_all_labels(self):
        labels = []
        for row in self.treeview_data_table.get_model():
            label_data = self.get_label_from_row(row)
            if label_data:
                labels.append(label_data)
        return labels

    def button_print_every_label_clicked(self, widget, data=None): \
            # pylint: disable=unused-argument
        self.print_labels(
            self.get_all_labels(),
            self.entry_label_type.get_text(),
            self.get_copies(),
        )

    def button_sheet_every_label_clicked(self, widget, data=None): \
            # pylint: disable=unused-argument
        self.sheet_labels(
            self.get_all_labels(),
            self.entry_label_type.get_text(),
            self.get_copies()
        )

    def button_1_clicked(self, widget, data=None): \
            # pylint: disable=unused-argument
        self.entry_copies.set_text('1')

    def button_2_clicked(self, widget, data=None): \
            # pylint: disable=unused-argument
        self.entry_copies.set_text('2')

    def button_3_clicked(self, widget, data=None): \
            # pylint: disable=unused-argument
        self.entry_copies.set_text('3')

    def button_4_clicked(self, widget, data=None): \
            # pylint: disable=unused-argument
        self.entry_copies.set_text('4')

    def get_number_of_records(self):
        model = self.treeview_data_table.get_model()
        return len(model) if model else 0

    def adjust_number_records(self):
        number_of_records = self.get_number_of_records()
        self.label_number_of_records.set_text(str(number_of_records))
        self.button_print_every_label.set_sensitive(
            number_of_records > 0 and
            self.get_selected_printer_config() is not None
        )
        self.button_sheet_every_label.set_sensitive(number_of_records > 0)

    def button_paste_clicked(self, widget, data=None): \
            # pylint: disable=unused-argument
        # print("button_paste_clicked - %s" % data)
        set_treeview_from_clipboard(self.treeview_data_table, with_header=False)
        self.adjust_number_records()

    def button_paste_with_header_clicked(self, widget, data=None): \
            # pylint: disable=unused-argument
        # print("button_paste_with_header_clicked - %s" % data)
        set_treeview_from_clipboard(self.treeview_data_table, with_header=True)
        self.adjust_number_records()

    def button_demo_clicked(self, widget, data=None): \
            # pylint: disable=unused-argument
        Gtk.Clipboard.get(
            Gdk.SELECTION_CLIPBOARD
        ).set_text(
            DEMO_WITH_HEADER,
            -1
        )
        self.button_paste_with_header.emit("clicked")

    def button_generated_clicked(self, widget, data=None): \
            # pylint: disable=unused-argument
        try:
            with open(LABELS_GENERATED_FILE, 'r') as yaml_file:
                data_log = load(yaml_file.read())
            records = self.get_records()
            if records is not None:
                data_log = data_log[-records:]
            # print('generated labels:\n', data_log)
            self.set_records(data_log)
        except IOError:
            pass  # silently pass, because log file does not exist yet

    def button_printed_clicked(self, widget, data=None): \
            # pylint: disable=unused-argument
        try:
            with open(LABELS_PRINTED_FILE, 'r') as yaml_file:
                data_log = load(yaml_file.read())
            records = self.get_records()
            if records is not None:
                data_log = data_log[-records:]
            # print('printed labels:\n', data_log)
            self.set_records(data_log)
        except IOError:
            pass  # silently pass, because log file does not exist yet

    def set_records(self, data_log):
        set_treeview_from_text_data(
            self.treeview_data_table,
            list_of_dicts_to_table(data_log),
            with_header=True,
        )
        self.adjust_number_records()

    def data_tree_selection_changed(self, widget, data=None): \
            # pylint: disable=unused-argument
        # print("selection_changed - %s" % data)
        csr = widget.count_selected_rows()
        # print("count_selected_rows:", csr)
        self.button_print_selected_labels.set_sensitive(
            csr is not None and
            (csr > 0) and
            self.get_selected_printer_config() is not None
        )
        self.button_sheet_selected_labels.set_sensitive(csr > 0)
        self.update_label_preview()

    def cb_printers_changed(self, widget, data=None):
        number_of_records = self.get_number_of_records()
        self.button_print_every_label.set_sensitive(
            number_of_records > 0 and self.get_selected_printer_config()
        )
        csr = self.treeview_data_table.get_selection().count_selected_rows()
        self.button_print_selected_labels.set_sensitive(
            csr is not None and
            (csr > 0) and
            self.get_selected_printer_config() is not None
        )

    def update_label_preview(self):
        if self.template in self.templates:
            self.box_preview.set_visible(True)
        else:
            self.box_preview.set_visible(False)
            return
        selected_labels = self.get_selected_labels()
        if selected_labels is not None and len(selected_labels) == 1:
            size_x, size_y = self.get_label_size()

            self.selected_label = selected_labels[0]
            self.selected_label_file = self.get_label_file(self.selected_label)
            self.textview.get_buffer().set_text(self.selected_label_file)
            self.drawingarea.set_size_request(
                self.dpi * size_x / 25.4,
                self.dpi * size_y / 25.4
            )
            self.drawingarea.queue_draw()
        else:
            self.textview.get_buffer().set_text('')
            self.drawingarea.set_size_request(-1, -1)
            self.selected_label = None
            self.selected_label_file = None

    def on_draw(self, wid, cr): \
            # pylint: disable=unused-argument
        if self.template not in self.templates:
            return
        # frame it!
        size_x, size_y = self.get_label_size()
        cr.set_line_width(1)
        cr.rectangle(
            0,
            0,
            self.dpi * size_x / 25.4,
            self.dpi * size_y / 25.4,
        )
        cr.set_source_rgb(255, 255, 255)
        cr.fill()
        # cr.set_source_rgb(0, 0, 0)
        # cr.stroke()
        draw_label_file(
            (
                self.selected_label_file
                if hasattr(self, 'selected_label_file') else None
            ),
            cr,
            self.dpi,
            self.get_image_file,
        )

    def get_image_file(self, name):
        for file_name, short_name in self.image_names.items():
            if name == short_name:
                local_file = join(svg_dir, file_name)
                if not isfile(local_file):
                    raise ImageFileMissing(local_file)
                return local_file
        if name in self.tempfiles:
            return self.tempfiles[name]
        raise ValueError('Unrecognized image name: %s' % name)

    def params_tree_selection_changed(self, widget, data=None):
        # print("selection_changed - %s" % data)
        # csr = widget.count_selected_rows()
        # print("params_tree_selection_changed count_selected_rows:", csr)
        pass

    def get_parameters(self):
        if self.template in self.templates:
            my_template = self.templates[self.template]
            if 'parameters' in my_template:
                return my_template['parameters']
        # it is possible to have a label without parameters
        return []

    def get_label_type(self):
        return self.label_types[
            self.templates[self.template]['label type']
        ]

    def get_label_size(self):
        label_type = self.get_label_type()
        return label_type['width'], label_type['height']

    def cb_part_numbers_selection_changed(self, widget, data=None):
        # print("cb_part_numbers_selection_changed - %s" % data)
        # reset a few entries
        self.document_name = None
        self.entry_document.set_text('')
        self.template = None
        self.entry_label_template.set_text('')
        self.entry_label_type.set_text('')
        self.qr_code_standard = None
        self.entry_qr_code_standard.set_text('')
        self.box_qr_code_standard.set_visible(False)
        set_treeview_from_texts(
            self.treeview_params,
            [],
        )
        # find out the selected part number
        tree_iter = widget.get_active_iter()
        if tree_iter is not None:
            sales_level_part_number = widget.get_model()[tree_iter][0]
            # TBD
            if sales_level_part_number in self.part_numbers:
                self.sales_level_part_number = sales_level_part_number
                part_number_data = self.part_numbers[
                    self.sales_level_part_number
                ]
                self.cb_part_numbers.set_tooltip_text(
                    part_number_data['Description']
                    if 'Description' in part_number_data else ''
                )
                if 'spd' in part_number_data:
                    spd = part_number_data['spd']
                    # print('spd:', spd)
                    if spd in self.spds:
                        self.spd = spd
                        spd_data = self.spds[spd]
                        self.items = (
                            spd_data['items'] if 'items' in spd_data else {}
                        )
                        spd_entry = spd
                        if 'Revision' in spd_data:
                            spd_entry += '_' + spd_data['Revision']
                        self.entry_spd.set_text(spd_entry)
                        self.entry_spd.set_tooltip_text(
                            spd_data['Description']
                            if 'Description' in spd_data else ''
                        )
                    else:
                        self.items = {}
                else:
                    self.spd = None
                    self.entry_spd.set_text('')
                    self.items = (
                        part_number_data['items']
                        if 'items' in part_number_data
                        else {}
                    )
                fill_combo(
                    self.cb_documents,
                    sorted(self.items.keys())
                )
        # self.update_label_preview()

    def cb_documents_selection_changed(self, widget, data=None):
        tree_iter = widget.get_active_iter()
        if tree_iter is not None:
            item = self.items[widget.get_model()[tree_iter][0]]
            self.document_name = (
                item['part number'] if 'part number' in item else ''
            )
            self.entry_document.set_text(
                self.document_name + '_' + item['Revision']
                if 'Revision' in item else self.document_name
            )
            self.entry_document.set_tooltip_text(
                item['Description'] if 'Description' in item else ''
            )
            if 'qr code standard' in item:
                self.qr_code_standard = item['qr code standard']
                self.entry_qr_code_standard.set_text(self.qr_code_standard)
                self.box_qr_code_standard.set_visible(True)
            else:
                self.qr_code_standard = None
                self.entry_qr_code_standard.set_text('')
                self.box_qr_code_standard.set_visible(False)
            self.template = (
                item['template'] if 'template' in item else self.document_name
            )
            try:
                document_file = self.get_file(self.document_name)
                self.button_view_document.set_sensitive(True)
                self.button_view_document.set_tooltip_text(document_file)
            except (MultipleFiles, MissingFile):
                self.button_view_document.set_sensitive(False)
                self.button_view_document.set_tooltip_text('')
            self.entry_label_type.set_text(
                self.templates[self.template]['label type']
                if self.template in self.templates else ''
            )
            self.entry_label_type.set_tooltip_text(
                dump(self.templates[self.template]['label type'], width=50)
                if self.template in self.templates else ''
            )
            self.entry_label_template.set_text(
                self.template if self.template
                else ''
            )
            set_treeview_from_texts(
                self.treeview_params,
                self.get_parameters()
            )
            self.update_label_preview()

    def button_today_clicked(self, widget, data=None): \
            # pylint: disable=unused-argument
        # print("button_today_clicked - %s" % data)
        self.set_date_code_to_today()

    def set_date_code_to_today(self):
        today = datetime.date.today()
        todays_week = today.strftime("%W")
        todays_year = today.strftime("%Y")
        todays_date_code = str(todays_year[-2:]) + str(todays_week)
        self.entry_date_code.set_text(todays_date_code)

    def get_label_file(self, label_params, x_shift=None, y_shift=None):
        for param in self.get_parameters():
            if param not in label_params:
                raise ValueError('label parameter %s is missing' % repr(param))
        label_format_params = {
            'sales level part number': self.sales_level_part_number,
        }
        label_format_params.update(label_params)
        if 'date code' in label_params:
            label_format_params['YY'] = label_params['date code'][:2]
            label_format_params['WW'] = label_params['date code'][2:]
        if 'snap mac address' in label_params:
            label_format_params[
                'snap node address'
            ] = label_params['snap mac address'][-6:]
        if 'ethernet mac address' in label_params:
            label_format_params[
                'ethernet mac address last 6'
            ] = label_params['ethernet mac address'][-6:]
        # print('label_format_params:', label_format_params)
        if self.template not in self.templates:
            print('not implemented label type: %s' % self.template)
            return None
        svg_config_data = self.templates[self.template]
        if 'method' in svg_config_data:
            method = svg_config_data['method']
            if method == 'bitmap':
                return shift_and_format_items(
                    self.get_label_bitmap(
                        self.template,
                        label_format_params
                    )['items'],
                    x_shift=x_shift,
                    y_shift=y_shift,
                    label_format_params=label_format_params
                )
        return shift_and_format_items(
            self.get_label(self.template)['items'],
            x_shift=x_shift,
            y_shift=y_shift,
            label_format_params=label_format_params
        )

    def get_image_downloads(self, label_file):
        image_downloads = []
        for line in label_file.splitlines():
            if line.startswith('I'):
                print('image line:', line)
                image_name = line.split(';')[1]
                image_file = self.get_image_file(image_name)
                bmp_file = image_file[:-3] + 'bmp'
                if not isfile(bmp_file):
                    with tempfile.NamedTemporaryFile(
                        prefix='',
                        suffix='.bmp',
                        delete=False
                    ) as temp_file:
                        bmp_file = temp_file.name
                    convert_png_to_bmp(
                        image_file,
                        bmp_file
                    )
                    # start_file(bmp_file)
                    print('bmp_file:', bmp_file)
                    with open(bmp_file, 'rb') as bmp_stream:
                        bmp_binary = bmp_stream.read().replace(
                            '\x1b', '\x1b\x1b'
                        )
                    os.remove(bmp_file)
                else:
                    with open(bmp_file, 'rb') as bmp_stream:
                        bmp_binary = bmp_stream.read().replace(
                            '\x1b', '\x1b\x1b'
                        )
                image_download = (
                    'd ' + 'BMP' + ';' + image_name + ' \r' +
                    '\x1b.' + bmp_binary + '\x1b.'
                )
                print(
                    'image_name:', image_name,
                    'image_file_path:', image_file,
                    'length:', len(image_download),
                )
                image_downloads.append(image_download)
        return image_downloads

    def get_selected_printer_name(self):
        tree_iter = self.cb_printers.get_active_iter()
        if tree_iter is not None:
            return self.cb_printers.get_model()[tree_iter][0]

    def get_selected_printer_config(self):
        selected_printer_name = self.get_selected_printer_name()
        if selected_printer_name is not None:
            return self.printers[selected_printer_name]

    def print_label_file(self, label_file, label_records):
        # print('print_label_file:\n"""\n%s"""' % label_file)
        # font_downloads = get_font_downloads()
        image_downloads = self.get_image_downloads(label_file)
        # label_file = '\n'.join(font_downloads) + label_file
        if len(image_downloads) > 0:
            label_file = '\r\n'.join(image_downloads) + '\r\n' + label_file
        selected_printer_config = self.get_selected_printer_config()
        if selected_printer_config is not None:
            try:
                print_label_file_to(label_file, selected_printer_config)
                self.save_print_records(label_records)
            except ftplib.all_errors:
                print("Could not print labels")
        else:
            print("No printer is selected")

    def save_print_records(self, labels):
        # labels = self.get_all_labels()
        time_before = None
        for label in labels:
            label['sales level part number'] = self.sales_level_part_number
            while True:
                time_now = now_string()
                if time_now != time_before:
                    time_before = time_now
                    break
            label['created'] = time_now
            selected_printer_name = self.get_selected_printer_name()
            if selected_printer_name is not None:
                label['printer'] = selected_printer_name
            if self.template is not None:
                label['template'] = self.template
            with open(LABELS_PRINTED_FILE, 'a') as yaml_file:
                yaml_file.write('- ' + dump(label, width=10000))

    def print_labels(self, labels, label_type_name, copies=1):
        print('print labels:', labels)
        label_type = self.label_types[label_type_name]
        x_shift = 0
        lpr = label_type['lpr']
        total_width = label_type['width']
        if lpr > 1:
            total_width += (lpr - 1) * label_type['hrepeat']
        # print('lpr:', lpr)
        label_file = None
        label_records = []
        if labels is not None:
            for label in labels:
                for copy in range(copies):  # pylint: disable=unused-variable
                    label_records.append(label)
                    current_label_file = self.get_label_file(
                        label,
                        (x_shift * label_type['hrepeat'])
                        if x_shift else None
                    )
                    if label_file is None:
                        # label_file = 'M l fnt;arial\n'
                        label_file = 'm m\r\n'
                        label_file += 'J\r\n'
                        # label_file = 'F 10;Arial\n'
                        # label_file = 'F 11;ARIALB\n'
                        # label_file = 'F 12;TREBU\n'
                        # label_file = 'F 13;TREBUB\n'
                        label_file += (
                            'S l1;0,0,' +
                            str(label_type['height']) + ',' +
                            str(label_type['vrepeat']) + ',' +
                            str(total_width) +
                            '\r\n'
                        )
                        label_file += 'O R\r\n'
                    label_file += current_label_file
                    x_shift += 1
                    if x_shift >= lpr:
                        label_file += 'A 1\r\n'
                        self.print_label_file(label_file, label_records)
                        label_file = None
                        label_records = []
                        x_shift = 0
        if label_file is not None:
            label_file += 'A 1\r\n'
            self.print_label_file(label_file, label_records)
            label_file = None
            label_records = []
            x_shift = 0

    def button_generate_clicked(self, widget, data=None): \
            # pylint: disable=unused-argument
        # print("button_generate_clicked - %s" % data)
        set_treeview_from_text_data(
            self.treeview_data_table,
            self.generate_records(),
            with_header=True,
        )
        self.adjust_number_records()
        self.save_records()

    def save_records(self):
        labels = self.get_all_labels()
        time_before = None
        for label in labels:
            label['sales level part number'] = self.sales_level_part_number
            while True:
                time_now = now_string()
                if time_now != time_before:
                    time_before = time_now
                    break
            label['created'] = time_now
            save_label(label)

    def get_records(self):
        try:
            return int(self.entry_records.get_text())
        except ValueError:
            return None

    def entry_records_changed(self, widget, data=None): \
            # pylint: disable=unused-argument
        records = self.get_records()
        self.button_generate.set_sensitive(records is not None and records > 0)

    def generate_records(self):
        try:
            records = int(self.entry_records.get_text())
        except ValueError:
            records = 0
        return self.generate_n_records(records)

    def sheet_labels(self, labels, label_type_name, copies=1):
        # typical paper and margin sizes
        # letter size: 8.5 x 11 in.
        # printable area on letter: 8.2 x 10.6 in.
        # top margin .22 in.
        # bottom margin: .18 in
        # sides margin .15 in
        label_type = self.label_types[label_type_name]
        label_x_mm = label_type['width']
        label_y_mm = label_type['height']
        paper_x_mm = 8.5 * 25.4
        paper_y_mm = 11 * 25.4
        side_margin_mm = 0.15 * 25.4
        top_margin_mm = 0.22 * 25.4
        bottom_margin_mm = 0.15 * 25.4
        printable_x_mm = paper_x_mm - side_margin_mm - side_margin_mm
        printable_y_mm = paper_y_mm - top_margin_mm - bottom_margin_mm
        labels_x = int(printable_x_mm / label_x_mm)
        labels_y = int(printable_y_mm / label_y_mm)
        # print(
        #     'labels_x:', labels_x, 'labels_y:', labels_y,
        #     'len:', len(labels)
        #     )
        # labels_page = labels_x * labels_y
        doc_dir = GLib.get_user_special_dir(
            GLib.UserDirectory.DIRECTORY_DOCUMENTS
        )
        if doc_dir is None:
            doc_dir = GLib.get_home_dir()
        pdf_file = join(doc_dir, 'gtk-demo.pdf')
        # print('pdf_file:', pdf_file)
        pdfdpi = 300.0
        ps = cairo.PDFSurface(
            pdf_file,
            pdfdpi * paper_x_mm / 25.4,
            pdfdpi * paper_y_mm / 25.4
        )
        cr = cairo.Context(ps)
        x_pos = 0
        y_pos = 0
        for label in labels:
            for copy in range(copies):  # pylint: disable=unused-variable
                # print('x_pos:', x_pos, 'y_pos:', y_pos)
                x_shift_mm = side_margin_mm + x_pos * label_x_mm
                y_shift_mm = top_margin_mm + y_pos * label_y_mm
                label_file = self.get_label_file(
                    label,
                    x_shift=x_shift_mm,
                    y_shift=y_shift_mm
                )
                # frame it!
                cr.set_line_width(1)
                cr.rectangle(
                    pdfdpi * x_shift_mm / 25.4,
                    pdfdpi * y_shift_mm / 25.4,
                    pdfdpi * label_x_mm / 25.4,
                    pdfdpi * label_y_mm / 25.4,
                )
                # cr.set_source_rgb(255, 255, 255)
                # cr.fill()
                cr.set_source_rgb(0, 0, 0)
                cr.stroke()
                draw_label_file(
                    label_file,
                    cr, pdfdpi,
                    self.get_image_file,
                )
                x_pos += 1
                if x_pos >= labels_x:
                    x_pos = 0
                    y_pos += 1
                    if y_pos >= labels_y:
                        y_pos = 0
                        cr.show_page()
        if x_pos > 0 or y_pos > 0:
            cr.show_page()
        ps.finish()
        if isfile(pdf_file):
            start_file(pdf_file)

    def button_view_spd_clicked(self, widget, data=None): \
            # pylint: disable=unused-argument
        if self.spd is not None:
            start_file(self.get_file(self.spd))

    def button_view_document_clicked(self, widget, data=None): \
            # pylint: disable=unused-argument
        if self.document_name is not None:
            start_file(self.get_file(self.document_name))

    def button_view_qr_code_standard_clicked(self, widget, data=None): \
            # pylint: disable=unused-argument
        if self.qr_code_standard is not None:
            start_file(self.get_file(self.qr_code_standard))


def synapse_labels():
    basicConfig(level=INFO)
    gui = LabelsGui()
    assert gui
    Gtk.main()
